from typing import Optional
from matplotlib import pyplot as plt
import numpy as np
from scipy.signal import find_peaks


from cut_detector.factories.mt_cut_detection_factory import (
    MtCutDetectionFactory,
)

from developers.binary_bridges_cnn.semi_images_generation.peak import Peak


class MtCutDetectionFactoryAdvanced(MtCutDetectionFactory):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Min ratio distance on circle between two peaks
        self.circle_min_ratio = 0.2025

    def _detect_peaks_from_intensities(
        self,
        intensities: list[float],
        circle_index: int,
        coeff_height_peak: float,
        coeff_width_peak: float,
        debug_plot: bool,
        circle_positions: Optional[list[tuple[int]]] = None,
    ) -> list[Peak]:
        """
        Given a list of intensities, detect peaks using scipy.

        Parameters
        ----------
        circle_positions : (y, x)
        """

        # Concatenate intensities to detect peaks even at borders
        concatenated_intensities = np.concatenate(
            (
                np.array(intensities),
                np.array(intensities),
                np.array(intensities),
            ),
            axis=0,
        )

        minimum_height = np.mean(intensities) * coeff_height_peak
        peaks_idx, peaks_data = find_peaks(
            concatenated_intensities,
            height=(
                minimum_height,
                None,
            ),  # absolute value
            distance=len(intensities) * self.circle_min_ratio,
            width=(None, round(len(intensities) * coeff_width_peak)),
            prominence=(None, None),
        )

        # Create peak instances
        # Regarding peak_idx, remove intensities length to focus on the first intensities list
        peaks = [
            Peak(
                relative_position=(peak_idx - len(intensities))
                / len(intensities),
                intensity=peaks_data["peak_heights"][idx],
                coordinates=(
                    circle_positions[peak_idx - len(intensities)]
                    if circle_positions is not None
                    else (0, 0)
                ),
                relative_intensity=peaks_data["peak_heights"][idx]
                / np.mean(intensities),
                position_index=peak_idx - len(intensities),
                circle_index=circle_index,
                prominence=peaks_data["prominences"][idx],
                width=peaks_data["widths"][idx],
                relative_width=peaks_data["widths"][idx] / len(intensities),
            )
            for idx, peak_idx in enumerate(peaks_idx)
            if len(intensities) <= peak_idx < 2 * len(intensities)
        ]

        if debug_plot:
            results_half = (
                peaks_data["widths"],
                peaks_data["width_heights"],
                peaks_data["left_ips"],
                peaks_data["right_ips"],
            )  # ie taken at half prominence
            plt.figure()
            plt.plot(concatenated_intensities)
            plt.plot(
                peaks_idx, np.array(concatenated_intensities)[peaks_idx], "x"
            )
            plt.hlines(*results_half[1:], color="C2")
            contour_heights = (
                np.array(concatenated_intensities)[peaks_idx]
                - peaks_data["prominences"]
            )
            plt.vlines(
                x=peaks_idx,
                ymin=contour_heights,
                ymax=concatenated_intensities[peaks_idx],
                color="green",
            )
            plt.xlim(
                len(intensities), len(intensities) * 2
            )  # ignore rest of concatenated intensities
            plt.show()

        return peaks

    def get_circle_data(
        self,
        radius: int,
        image: np.ndarray,
    ) -> tuple[list[tuple[int]], list[float], list[Peak]]:
        """
        Compute useful data for a circle of a given radius around the mid body spot:

        Parameters
        -------
        image : YX

        Returns
        -------

        intensities : list of the intensities of the pixels on the circle

        """
        # Get image shape
        size_y, size_x = image.shape[0], image.shape[1]

        # Get the mid-body spot coordinates, middle of the image
        mid_body_position = (size_x // 2, size_y // 2)

        # Get associated perimeter, angles and coordinates
        perimeter = round(2 * np.pi * radius)
        angles = np.linspace(0, 2 * np.pi, perimeter)
        circle_positions = []
        for angle in angles:
            pos_y = round(mid_body_position[0] + radius * np.cos(angle))
            pos_x = round(mid_body_position[1] + radius * np.sin(angle))
            if (pos_y, pos_x) not in circle_positions:
                circle_positions.append((pos_y, pos_x))

        intensities = []
        for position in circle_positions:
            # Get the mean intensity around the mid spot
            total, inc = 0, 0
            for k in range(-1, 2):
                for j in range(-1, 2):
                    pos_y = position[0] + k
                    pos_x = position[1] + j
                    if 0 <= pos_y < size_y and 0 <= pos_x < size_x:
                        total += image[pos_y, pos_x]
                        inc += 1
            mean = total / inc
            intensities.append(mean)

        return intensities

    def get_average_circle_peaks(
        self, all_intensities: list[float], debug_plot: bool
    ) -> list[Peak]:
        """
        Compute 2 best peaks on average circle.
        """
        # First, compute average circle intensities
        expected_points_nb = len(all_intensities[0])
        average_circle_intensities = []
        for intensities in all_intensities:
            interpolated_points = np.linspace(
                0, len(intensities), expected_points_nb
            )
            interpolated_intensities = np.interp(
                interpolated_points,  # x
                range(len(intensities)),  # xp
                intensities,  # fp
            )
            average_circle_intensities.append(interpolated_intensities)
        average_circle_intensities = np.mean(
            average_circle_intensities, axis=0
        )

        # Then, detect peaks on average circle
        average_circle_peaks = self._detect_peaks_from_intensities(
            average_circle_intensities,
            circle_index=0,
            coeff_height_peak=0,  # minimum
            coeff_width_peak=len(average_circle_intensities),  # maximum
            debug_plot=debug_plot,
        )

        def get_peak_prominence(peak: Peak):
            return peak.prominence

        # Keep only 2 best peaks, i.e. with highest prominence
        average_circle_peaks.sort(key=get_peak_prominence, reverse=True)
        average_circle_peaks = average_circle_peaks[:2]

        # Complete with empty Peak if length <2
        while len(average_circle_peaks) < 2:
            average_circle_peaks.append(Peak())

        return average_circle_peaks



import torch

from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)
from cnn_framework.utils.data_loader_generators.classifier_data_loader_generator import (
    ClassifierDataLoaderGenerator,
)
from cnn_framework.utils.metrics.classification_accuracy import (
    ClassificationAccuracy,
)
from cnn_framework.utils.models.resnet_classifier import ResnetClassifier

from developers.binary_bridges_cnn.data_set import BinaryBridgesCnnDataSet
from developers.binary_bridges_cnn.model_params import BinaryBridgesModelParams
from developers.binary_bridges_cnn.bridges_parser import BridgesParser


def main(params):
    loader_generator = ClassifierDataLoaderGenerator(
        params, BinaryBridgesCnnDataSet
    )
    _, _, test_dl = loader_generator.generate_data_loader()

    # Model definition
    model = ResnetClassifier(
        nb_classes=2,  # hard-coded from binary training
        nb_input_channels=len(params.c_indexes) * len(params.z_indexes),
        encoder_name=params.encoder_name,
    )
    model.load_state_dict(torch.load(params.model_load_path))

    manager = CnnModelManager(model, params, ClassificationAccuracy)
    manager.predict(test_dl, nb_images_to_save=-1)

    manager.write_useful_information()


if __name__ == "__main__":
    parser = BridgesParser()
    args = parser.arguments_parser.parse_args()

    parameters = BinaryBridgesModelParams()
    parameters.update(args)

    main(parameters)



from __future__ import annotations

import numpy as np
import pandas as pd

from .track import Track
from .mid_body_spot import MidBodySpot


class MidBodyTrack(Track[MidBodySpot]):
    """
    Mid-body candidate track
    """

    @staticmethod
    def track_df_to_track_list(
        track_df: pd.DataFrame,
        spots: dict[int, list[MidBodySpot]],
    ) -> list[MidBodyTrack]:

        track_df.reset_index(inplace=True)
        track_df.dropna(inplace=True)
        id_to_track = {}

        for _, row in track_df.iterrows():
            track_id = row["track_id"]
            track: MidBodyTrack = id_to_track.get(track_id)
            if track is None:
                id_to_track[track_id] = MidBodyTrack(len(id_to_track))
                track = id_to_track[track_id]
            frame = row["frame"]
            idx_in_frame = row["idx_in_frame"]
            track.add_spot(spots[int(frame)][int(idx_in_frame)])

        return list(id_to_track.values())

    def get_expected_distance(
        self,
        expected_positions: dict[int, list[int]],
        max_distance: float,
    ) -> float:
        """
        Compute the average distance between mid-body expected positions and current
        track positions.
        """
        distances = []
        for frame, position in expected_positions.items():
            if frame not in self.spots:
                continue
            spot = self.spots[frame]
            distances.append(
                np.linalg.norm(spot.get_position() - np.array(position))
            )
        # If there are no frames in common, for sure track is not the right one
        if len(distances) == 0:
            return np.inf
        mean_distance = np.mean(distances)
        # If the mean distance is too high, discard the track
        if mean_distance > max_distance:
            return np.inf
        return mean_distance

    def fill_gaps(self):
        """
        Fill gaps in the track.
        """
        frames = list(self.spots.keys())
        min_frame, max_frame = min(frames), max(frames)
        for frame in range(min_frame, max_frame):
            next_frame = frame + 1
            # Look for the next frame with a spot
            while next_frame not in self.spots:
                next_frame += 1
            # If it was not the next frame, fill the gap
            if next_frame != frame + 1:
                gap_size = next_frame - (frame + 1)  # number of missing spots
                current_spot = self.spots[frame]
                next_spot = self.spots[next_frame]
                # Define interpolated values
                ranges = {}
                for attribute in [
                    "x",
                    "y",
                    "intensity",
                    "sir_intensity",
                    "area",
                    "circularity",
                ]:
                    if getattr(current_spot, attribute) is None:
                        ranges[attribute] = [None] * (gap_size + 2)
                    else:
                        ranges[attribute] = np.linspace(
                            getattr(current_spot, attribute),
                            getattr(next_spot, attribute),
                            gap_size + 2,
                        )
                for i in range(1, gap_size + 1):
                    new_spot = MidBodySpot(
                        frame=frame + i,
                        x=int(ranges["x"][i]),
                        y=int(ranges["y"][i]),
                        intensity=ranges["intensity"][i],
                        sir_intensity=ranges["sir_intensity"][i],
                        area=ranges["area"][i],
                        circularity=ranges["circularity"][i],
                    )
                    self.add_spot(new_spot)
        # All gaps should have been filled
        assert self.length == max_frame - min_frame + 1



import os
import random
import xml.etree.ElementTree as ET
from bigfish import stack


def extract_images(
    xml_model_path: str,
    video_path: str,
    save_dir: str,
    ratio_to_extract: float,
    circularity_threshold: float,
) -> None:
    """Extract images from a video based on a CellCounter xml file.

    Parameters
    ----------
    xml_model_path : str
        Path to the xml file.
    video_path : str
        Path to the video.
    save_dir : str
        Path to the folder where to save the images.
    ratio_to_extract : float
        Ratio of images to extract.
    circularity_threshold : float
        Circularity threshold to extract images.

    """
    root = ET.parse(xml_model_path).getroot()

    # Create save directory
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # Read image
    raw_video = stack.read_image(video_path, sanity_check=False)  # T, H, W, C
    video_name = os.path.basename(video_path).split(".")[0]

    for type_tag in root.findall("Model/AllSpots/SpotsInFrame/Spot"):
        raw_positions = type_tag.text.split(" ")
        frame = int(type_tag.get("FRAME"))

        # Get min and max positions
        rel_positions_x = [
            float(raw_positions[i]) for i in range(0, len(raw_positions), 2)
        ]
        rel_positions_y = [
            float(raw_positions[i]) for i in range(1, len(raw_positions), 2)
        ]
        rel_min_x, rel_max_x = min(rel_positions_x), max(rel_positions_x)
        rel_min_y, rel_max_y = min(rel_positions_y), max(rel_positions_y)

        spot_position_x, spot_position_y = (
            float(type_tag.get("POSITION_X")),
            float(type_tag.get("POSITION_Y")),
        )
        abs_min_x, abs_max_x = int(spot_position_x + rel_min_x), int(
            spot_position_x + rel_max_x
        )
        abs_min_y, abs_max_y = int(spot_position_y + rel_min_y), int(
            spot_position_y + rel_max_y
        )

        abs_min_x, abs_max_x = max(abs_min_x, 0), min(
            abs_max_x, raw_video.shape[2]
        )
        abs_min_y, abs_max_y = max(abs_min_y, 0), min(
            abs_max_y, raw_video.shape[1]
        )

        nucleus = raw_video[frame, abs_min_y:abs_max_y, abs_min_x:abs_max_x, :]

        circularity = float(type_tag.get("CIRCULARITY"))

        if (
            circularity
            > circularity_threshold  # Extract rounded nuclei to have metaphase cells
            or random.random() < ratio_to_extract
            or frame
            > raw_video.shape[0]
            * 0.9  # Extract all last 10% of frames to have dead cells
        ):
            # Generate random index number
            random_index = random.randint(0, 100)
            stack.save_image(
                nucleus,
                f"{save_dir}/{video_name}_nucleus_{random_index}_{frame}.tif",
            )


if __name__ == "__main__":

    XML_MODEL_PATH = ""
    VIDEO_PATH = ""
    SAVE_DIR = ""

    RATIO_TO_EXTRACT = 0.01
    CIRCULARITY_THRESHOLD = 0.8

    extract_images(
        XML_MODEL_PATH,
        VIDEO_PATH,
        SAVE_DIR,
        RATIO_TO_EXTRACT,
        CIRCULARITY_THRESHOLD,
    )



"""Playground to run mid-body detection on a single mitosis.
Detected spots and tracks are printed in the console."""

import os
from typing import Optional
import pickle

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.factories.mid_body_detection_factory import (
    MidBodyDetectionFactory,
)
from cut_detector.utils.mid_body_detection.tracking import TRACKING_FUNCTIONS
from cut_detector.utils.mitosis_track import MitosisTrack
from cut_detector.utils.tools import re_organize_channels
from cut_detector.utils.track_generation import generate_tracks_from_spots


def main(
    image_path: Optional[str] = os.path.join(
        get_data_path("mitosis_movies"), "example_video_mitosis_0_0_to_5.tiff"
    ),
    mitosis_path: Optional[str] = get_data_path("mitoses"),
    show_points: bool = True,
    show_tracks: bool = True,
) -> None:
    """
    Parameters
    ----------
    image_path : str
        Path to the image to process.
    mitosis_path : str
        Path to the mitoses data.
    show_points : bool
        If True, plot the detected spots.
    show_tracks : bool
        If True, plot the detected tracks.
    """

    # If paths are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])
    if os.path.isdir(mitosis_path):
        mitosis_path = os.path.join(mitosis_path, os.listdir(mitosis_path)[0])

    # Read data: image, mitosis_track
    image = TiffReader(image_path, respect_initial_type=True).image  # TCZYX
    with open(mitosis_path, "rb") as f:
        mitosis_track: MitosisTrack = pickle.load(f)

    mitosis_movie = image[:, :3, ...].squeeze()  # T C=3 YX
    mitosis_movie = re_organize_channels(mitosis_movie)  # TYXC

    # Search for mid-body in mitosis movie
    factory = MidBodyDetectionFactory()

    spots_candidates = factory.detect_mid_body_spots(
        mitosis_movie=mitosis_movie,
        method="difference_gaussian",
        parallelization=False,
        mitosis_track=mitosis_track,
    )

    if show_points:
        print("\nSpots candidates:")
        for frame, spots in spots_candidates.items():
            for spot in spots:
                print(
                    {
                        "fr": frame,
                        "x": spot.x,
                        "y": spot.y,
                        "mklp_int": spot.intensity,
                        "sir_int": spot.sir_intensity,
                    }
                )

    generate_tracks_from_spots(
        spots_candidates, TRACKING_FUNCTIONS["spatial_laptrack"]
    )

    if show_tracks:
        print("\nSpots candidates with tracks:")
        for frame, spots in spots_candidates.items():
            for spot in spots:
                print(
                    {
                        "fr": frame,
                        "x": spot.x,
                        "y": spot.y,
                        "mklp_int": spot.intensity,
                        "sir_int": spot.sir_intensity,
                        "track_id": spot.track_id,
                    }
                )


if __name__ == "__main__":
    main()



import os
import xmltodict
from bigfish import stack
from munch import Munch
from cnn_framework.utils.tools import save_tiff

from cut_detector.factories.mt_cut_detection_factory import (
    MtCutDetectionFactory,
)
from cut_detector.utils.image_tools import (
    cell_counter_frame_to_video_frame,
    smart_cropping,
)
from cut_detector.constants.annotations import NAMES_DICTIONARY
from cut_detector.utils.tools import get_video_path


def extract_bridge_images(
    xml_folders, division_movies_path, save_dir, classes_to_keep
) -> None:
    """Extract bridge images from xml files.

    Parameters
    ----------
    xml_folders : list[str]
        List of paths to the folders containing xml CellCounter files.
    division_movies_path : str
        Path to the division movies folder.
    save_dir : str
        Path to the folder where to save the images.
    """
    margin = MtCutDetectionFactory().margin

    # Create save_dir if not exists
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    for xml_folder in xml_folders:
        for xml_name in os.listdir(xml_folder):
            print(f"Processing {xml_name}...")
            xml_path = os.path.join(xml_folder, xml_name)

            # Read useful information from xml file
            with open(xml_path) as fd:
                doc = Munch.fromDict(xmltodict.parse(fd.read()))

            # Read video
            video_name = (
                doc.CellCounter_Marker_File.Image_Properties.Image_Filename
            )
            video_path = get_video_path(video_name, division_movies_path)
            if video_path is None:
                print(f"Video {video_name} not found. Skipping.")
                continue

            raw_video = stack.read_image(
                video_path, sanity_check=False
            )  # TYXC
            video_save_name = os.path.basename(video_path).split(".")[0]
            # Switch channels to beginning
            raw_video = raw_video.transpose(3, 0, 1, 2)  # CYXW
            nb_channels = raw_video.shape[0]

            classes_max_frames = [-1] * 5
            for i, type_data in enumerate(
                doc.CellCounter_Marker_File.Marker_Data.Marker_Type
            ):
                assert i == int(type_data.Type) - 1  # order should be kept
                class_index = NAMES_DICTIONARY[type_data.Name]
                # Ignore if no data
                if "Marker" not in type_data:
                    continue
                # Ignore if current class as to be ignored
                if class_index not in classes_to_keep:
                    continue
                markers = type_data.Marker
                if not isinstance(markers, list):
                    markers = [markers]
                for marker in markers:
                    x_pos, y_pos, frame = (
                        int(marker.MarkerX),
                        int(marker.MarkerY),
                        cell_counter_frame_to_video_frame(
                            int(marker.MarkerZ), nb_channels
                        ),
                    )
                    # Update max frame
                    classes_max_frames[(i + 1) % 5] = max(
                        classes_max_frames[(i + 1) % 5], frame
                    )
                    assert frame > classes_max_frames[(i + 1) % 5 - 1]
                    # Extract image
                    frame_image = raw_video[:, frame, :, :].squeeze()  # CYX
                    crop = smart_cropping(
                        frame_image, margin, x_pos, y_pos, pad=True
                    )  # CYX
                    # Define name
                    image_name = f"{video_save_name}_{str(frame).zfill(3)}_c{class_index}.tif"
                    save_tiff(
                        crop,
                        os.path.join(save_dir, image_name),
                        original_order="CYX",
                    )


if __name__ == "__main__":
    XML_FOLDERS = [""]
    DIVISION_MOVIES_PATH = ""
    SAVE_DIR = ""

    CLASSES_TO_KEEP = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    extract_bridge_images(
        XML_FOLDERS, DIVISION_MOVIES_PATH, SAVE_DIR, CLASSES_TO_KEEP
    )



from cnn_framework.utils.parsers.transfer_parser import TransferParser


class BridgesParser(TransferParser):
    """
    Parsing class for bridges inference.
    """

    def __init__(self):
        super().__init__()

        self.arguments_parser.add_argument(
            "--hmm_bridges_parameters_file",
            help="Path to bridges HMM parameters file.",
        )



import numpy as np
import torch
from torch.utils.data import DataLoader

from cnn_framework.utils.metrics.abstract_metric import AbstractMetric
from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)

from .micro_tubules_augmentation import (
    MicroTubulesAugmentation,
)


class BridgesMtModelManager(CnnModelManager):
    """Model manager for Binary Bridges CNN classification models."""

    def augment_image(self, input_tensor: torch.Tensor) -> torch.Tensor:
        """Function to augment image following binary bridges principle.

        Parameters
        ----------
        input_tensor : torch.Tensor
            Input tensor to augment.

        Returns
        -------
        torch.Tensor
            Augmented tensor.
        """
        input_arrays = input_tensor.detach().cpu().numpy()  # BCYX

        augmentation_tool = MicroTubulesAugmentation()

        augmented_array = []
        for input_array in input_arrays:
            # Put first dimension last
            augmented_images = augmentation_tool.generate_augmentations(
                input_array
            )
            for augmented_dict in augmented_images.values():
                # Put last dimension first
                augmented_image = augmented_dict["image"]  # YXC
                augmented_image = augmented_image.transpose(2, 0, 1)  # CYX
                augmented_array.append(augmented_image)

        augmented_array = np.array(augmented_array)
        augmented_tensor = torch.from_numpy(augmented_array)
        return augmented_tensor

    def model_prediction(
        self,
        dl_element: DatasetOutput,
        dl_metric: AbstractMetric,
        data_loader: DataLoader,
    ) -> None:
        """Function to generate outputs from inputs for given model.

        Parameters
        ----------
        dl_element : DatasetOutput
            Element from the data loader.
        dl_metric : AbstractMetric
            Metric to update.
        data_loader : DataLoader
            Data loader.
        """
        augmented_input = self.augment_image(dl_element.input)
        augmentation_nb = augmented_input.shape[0] // dl_element.input.shape[0]

        dl_element.to_device(self.device)
        augmented_input = augmented_input.to(self.device)

        predictions = torch.softmax(
            self.model(augmented_input.float()), dim=-1
        )

        # From MT detection to cut detection
        predictions = predictions.view(
            -1, augmentation_nb // 2, 2, 2
        )  # augmentation_nb images, 2 scores
        final_predictions = torch.zeros_like(dl_element.target)
        for idx, image_predictions in enumerate(predictions):
            for pair_predictions in image_predictions:
                # NB: no cut means 2 MT
                no_cut_prediction = (
                    pair_predictions[0][1] * pair_predictions[1][1]
                )
                cut_prediction = (
                    pair_predictions[0][1] * pair_predictions[1][0]
                    + pair_predictions[0][0] * pair_predictions[1][1]
                )
                two_cuts_prediction = (
                    pair_predictions[0][0] * pair_predictions[1][0]
                )
                final_predictions[idx][0] += no_cut_prediction
                final_predictions[idx][1] += cut_prediction
                final_predictions[idx][2] += two_cuts_prediction
        dl_element.prediction = final_predictions

        # Update metric
        dl_metric.update(
            final_predictions,
            dl_element.target,
            adds=dl_element.additional,
            mean_std=data_loader.dataset.mean_std,
        )



"""Define optimal detection functions for every strategy."""

from functools import partial
import numpy as np
from skimage.feature import blob_log, blob_dog, blob_doh


def min_max(img: np.ndarray) -> np.ndarray:
    """Normalize the image between 0 and 1.

    Parameters
    ----------
    img : np.ndarray
        Image to normalize.

    Returns
    -------
    np.ndarray
        Normalized image.
    """
    img_max = np.max(img)
    img_min = np.min(img)
    return (img - img_min) / (img_max - img_min)


def detect_minmax_log(img: np.ndarray, **kwargs) -> np.ndarray:
    """Detect blobs using Laplacian of Gaussian with min-max normalization.

    Parameters
    ----------
    img : np.ndarray
        Image to detect blobs.
    **kwargs
        Extra arguments for blob_log.

    Returns
    -------
    np.ndarray
        Blobs detected."""
    img_norm = min_max(img)
    return blob_log(image=img_norm, **kwargs)


def detect_minmax_dog(img: np.ndarray, **kwargs) -> np.ndarray:
    """Detect blobs using Difference of Gaussian with min-max normalization.

    Parameters
    ----------
    img : np.ndarray
        Image to detect blobs.
    **kwargs
        Extra arguments for blob_dog.

    Returns
    -------
    np.ndarray
        Blobs detected."""
    img_norm = min_max(img)
    return blob_dog(img_norm, **kwargs)


def detect_minmax_doh(img: np.ndarray, **kwargs) -> np.ndarray:
    """Detect blobs using Determinant of Hessian with min-max normalization.

    Parameters
    ----------
    img : np.ndarray
        Image to detect blobs.
    **kwargs
        Extra arguments for blob_doh.

    Returns
    -------
    np.ndarray
        Blobs detected."""
    img_norm = min_max(img)
    return blob_doh(img_norm, **kwargs)


DETECTION_FUNCTIONS = {
    # Laplacian of Gaussian
    "laplacian_gaussian": partial(
        detect_minmax_log,
        min_sigma=5,
        max_sigma=10,
        num_sigma=5,
        threshold=0.1,
    ),
    "log2_wider": partial(
        detect_minmax_log, min_sigma=2, max_sigma=8, num_sigma=4, threshold=0.1
    ),
    "rshift_log": partial(
        detect_minmax_log,
        min_sigma=3,
        max_sigma=11,
        num_sigma=5,
        threshold=0.1,
    ),
    # Difference of Gaussian
    "difference_gaussian": partial(
        detect_minmax_dog,
        min_sigma=2,
        max_sigma=5,
        sigma_ratio=1.2,
        threshold=0.1,
    ),
    "very_fast_dog": partial(
        detect_minmax_dog,
        min_sigma=5,
        max_sigma=5,
        sigma_ratio=1.2,
        threshold=0.1,
    ),
    "dog_005": partial(
        detect_minmax_dog,
        min_sigma=2,
        max_sigma=5,
        sigma_ratio=1.2,
        threshold=0.05,
    ),
    # Determinant of Hessian
    "determinant_hessian": partial(
        detect_minmax_doh,
        min_sigma=5,
        max_sigma=10,
        num_sigma=5,
        threshold=0.0040,
    ),
}



from __future__ import annotations

import numpy as np


class Peak:
    """
    A class to store useful values regarding signal peaks.
    Used in the cut detection algorithm.
    """

    def __init__(
        self,
        relative_position=-1,
        intensity=0,
        coordinates=(0, 0),
        relative_intensity=0,
        position_index=-1,
        circle_index=-1,
        prominence=0,
        width=0,
        relative_width=0,
    ):
        self.relative_position = relative_position
        self.intensity = intensity
        self.coordinates = coordinates  # (y, x)
        self.relative_intensity = relative_intensity
        self.position_index = position_index
        self.circle_index = circle_index
        self.prominence = prominence
        self.width = width
        self.relative_width = relative_width

    def is_empty(self) -> bool:
        """
        Return True if the peak is empty, i.e. relative_position == -1.
        """
        return self.relative_position == -1

    @classmethod
    def get_average_intensity(cls, peaks: list[Peak]) -> float:
        """
        Return the average intensity of a list of peaks.
        """
        intensities = [peak.intensity for peak in peaks]
        return np.mean(intensities)

    @classmethod
    def get_average_relative_position(cls, peaks: list[Peak]) -> float:
        """
        Return the average relative position of a list of peaks.
        """
        relative_positions = [peak.relative_position for peak in peaks]
        return np.mean(relative_positions)

    @classmethod
    def get_average_relative_intensity(cls, peaks: list[Peak]) -> float:
        """
        Return the average relative intensity of a list of peaks.
        """
        relative_intensities = [peak.relative_intensity for peak in peaks]
        return np.mean(relative_intensities)

    @classmethod
    def get_average_prominence(cls, peaks: list[Peak]) -> float:
        """
        Return the average prominence of a list of peaks.
        """
        prominences = [peak.prominence for peak in peaks]
        return np.mean(prominences)

    @classmethod
    def get_average_width(cls, peaks: list[Peak]) -> float:
        """
        Return the average width of a list of peaks.
        """
        widths = [peak.width for peak in peaks]
        return np.mean(widths)

    @classmethod
    def get_average_relative_width(cls, peaks: list[Peak]) -> float:
        """
        Return the average relative width of a list of peaks.
        """
        relative_widths = [peak.relative_width for peak in peaks]
        return np.mean(relative_widths)

    @classmethod
    def group_peaks(
        cls,
        circle_peaks: list[list[Peak]],
        windows: list[tuple[float]],
        minimum_size: int,
    ) -> list[list[Peak]]:
        """
        Group the peaks by windows.
        """
        window_peaks = []
        for window in windows:
            group = []
            for peaks in circle_peaks:
                for peak in peaks:
                    if window[0] <= peak.relative_position < window[1]:
                        group.append(peak)
            window_peaks.append(group)

        # Filter out too small groups
        filtered_window_peaks = []
        for group_peak in window_peaks:
            if (
                len(group_peak) >= minimum_size
                and group_peak not in filtered_window_peaks
            ):
                filtered_window_peaks.append(group_peak)

        return filtered_window_peaks

    @classmethod
    def create_average_peak(cls, peaks: list[Peak]) -> Peak:
        """
        Create a peak with average values from a list of peaks.
        """
        average_relative_position = cls.get_average_relative_position(peaks)
        average_intensity = cls.get_average_intensity(peaks)
        average_coordinates = np.mean(
            [peak.coordinates for peak in peaks], axis=0
        )
        average_relative_intensity = cls.get_average_relative_intensity(peaks)
        average_prominence = cls.get_average_prominence(peaks)
        average_width = cls.get_average_width(peaks)
        average_relative_width = cls.get_average_relative_width(peaks)
        return cls(
            relative_position=average_relative_position,
            intensity=average_intensity,
            coordinates=average_coordinates,
            relative_intensity=average_relative_intensity,
            prominence=average_prominence,
            width=average_width,
            relative_width=average_relative_width,
        )

    @classmethod
    def remove_small_groups(
        cls, window_peaks: list[Peak], min_peaks_by_group: int
    ) -> list[Peak]:
        """
        Remove groups that have less than min_peaks_by_group peaks.
        """
        filtered_window_peaks = [
            group for group in window_peaks if len(group) >= min_peaks_by_group
        ]
        return filtered_window_peaks

    @classmethod
    def remove_duplicated_peaks(
        cls, window_peaks: list[Peak], min_peaks_by_group: int
    ) -> list[Peak]:
        """
        One peak must appear in one group only.
        Remove the peaks that are already seen in previous groups.
        Remove groups that have less than min_peaks_by_group peaks.
        """
        for i in range(len(window_peaks) - 1):
            for j in range(i + 1, len(window_peaks)):
                window_peaks[j] = [
                    p for p in window_peaks[j] if p not in window_peaks[i]
                ]
        filtered_window_peaks = cls.remove_small_groups(
            window_peaks, min_peaks_by_group
        )
        return filtered_window_peaks

    @classmethod
    def remove_close_peaks(
        cls, window_peaks: list[Peak], circle_min_ratio
    ) -> list[Peak]:
        """
        Remove the groups that are too close to each other, i.e. distance < circle_min_ratio.
        Choose the one with highest mean intensity (used to be highest number of peaks).
        """
        # Compute the mean position of the peaks in each group
        position_groups = [
            cls.get_average_relative_position(group) for group in window_peaks
        ]

        i = 0
        while i < len(position_groups) - 1:
            j = i + 1
            while j < len(position_groups):
                first_position = (position_groups[i] - position_groups[j]) % 1
                second_position = (position_groups[j] - position_groups[i]) % 1
                if min(first_position, second_position) < circle_min_ratio:
                    if cls.get_average_intensity(
                        window_peaks[i]
                    ) > cls.get_average_intensity(window_peaks[j]):
                        window_peaks.remove(window_peaks[j])
                        position_groups.remove(position_groups[j])
                        j -= 1
                    else:
                        window_peaks.remove(window_peaks[i])
                        position_groups.remove(position_groups[i])
                        i -= 1
                j += 1
            i += 1

        return window_peaks

    def intersects(self, relative_position: float) -> bool:
        """
        Return True if the peak intersects the given relative position.
        """
        relative_difference = min(
            abs(self.relative_position - relative_position),  # usual case
            1 - abs(self.relative_position - relative_position),  # wrap around
        )
        return relative_difference < self.relative_width / 2

    def enabled_augmentation(self) -> dict[str, int]:
        """
        Enable image augmentation.

        NB: for some reason, zero is at the bottom.
        Trigonometric rotation, so 0.25 is at the right, etc.
        """

        augmentations = {}

        # Top/bottom
        if self.intersects(0.25) or self.intersects(0.75):
            augmentations["top"] = -1
            augmentations["bottom"] = -1
        else:
            if (  # MT in top
                self.relative_position > 0.25 and self.relative_position < 0.75
            ):
                augmentations["top"] = 1
                augmentations["bottom"] = 0
            else:  # MT in left
                augmentations["top"] = 0
                augmentations["bottom"] = 1

        # Left/right
        if self.intersects(0.0) or self.intersects(0.5):
            augmentations["left"] = -1
            augmentations["right"] = -1
        else:
            if self.relative_position > 0.5:  # MT in left
                augmentations["left"] = 1
                augmentations["right"] = 0
            else:  # MT in right
                augmentations["left"] = 0
                augmentations["right"] = 1

        # Top right/bottom left
        if self.intersects(1 / 8) or self.intersects(5 / 8):
            augmentations["top_right"] = -1
            augmentations["bottom_left"] = -1
        else:
            if (  # MT in top right
                self.relative_position > 1 / 8
                and self.relative_position < 5 / 8
            ):
                augmentations["top_right"] = 1
                augmentations["bottom_left"] = 0
            else:  # MT in bottom left
                augmentations["top_right"] = 0
                augmentations["bottom_left"] = 1

        # Bottom right/top left
        if self.intersects(3 / 8) or self.intersects(7 / 8):
            augmentations["bottom_right"] = -1
            augmentations["top_left"] = -1
        else:
            if (  # MT in bottom right
                self.relative_position < 3 / 8
                or self.relative_position > 7 / 8
            ):
                augmentations["bottom_right"] = 1
                augmentations["top_left"] = 0
            else:  # MT in top left
                augmentations["bottom_right"] = 0
                augmentations["top_left"] = 1

        return augmentations



import os
from typing import Optional, Union
import pickle
import numpy as np
from tqdm import tqdm

from ..utils.cell_track import CellTrack
from ..utils.cell_spot import CellSpot
from ..factories.mitosis_track_generation_factory import (
    MitosisTrackGenerationFactory,
)
from ..models.tools import get_model_path
from ..utils.mitosis_track import MitosisTrack


def perform_mitosis_track_generation(
    video: np.ndarray,
    video_name: str,
    spots_dir: str,
    tracks_dir: str,
    mitoses_dir: Optional[str] = None,
    metaphase_model_path: Optional[str] = get_model_path("metaphase_cnn"),
    hmm_metaphase_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_metaphase_parameters.npz"
    ),
    predictions_file: Optional[str] = None,
    only_predictions_update: bool = False,
    save: bool = True,
) -> tuple[Union[list[MitosisTrack], None], list[CellSpot], list[CellTrack]]:
    """Perform mitosis track generation.

    Parameters
    ----------
    video : np.ndarray
        Video. TYXC.
    video_name : str
        Video name.
    spots_dir : str
        Directory where spots are saved.
    tracks_dir : str
        Directory where tracks are saved.
    mitoses_dir : Optional[str], optional
        Directory where mitoses are saved, by default None.
    metaphase_model_path : Optional[str], optional
        Metaphase model path, by default get_model_path("metaphase_cnn").
    hmm_metaphase_parameters_file : Optional[str], optional
        HMM metaphase parameters file, by default os.path.join(get_model_path("hmm"), "hmm_metaphase_parameters.npz").
    predictions_file : Optional[str], optional
        Predictions file, by default None.
    only_predictions_update : bool, optional
        Only update predictions, by default False.
    save : bool, optional
        Save, by default True.

    Returns
    -------
    Union[list[MitosisTrack], None]
        List of mitosis tracks if only_predictions_update is False, None otherwise.
    list[CellSpot]
        Cell spots.
    list[CellTrack]
        Cell tracks.
    """

    print("\n### CELL DIVISION DETECTION ###")

    # Create save_dir if not exists
    if mitoses_dir is not None and not os.path.exists(mitoses_dir):
        os.makedirs(mitoses_dir)

    # Create factory instance, where useful functions are defined
    tracks_merging_factory = MitosisTrackGenerationFactory()

    # Load cell spots
    cell_spots: list[CellSpot] = []
    video_spots_save_dir = os.path.join(spots_dir, video_name)
    for state_path in os.listdir(video_spots_save_dir):
        with open(os.path.join(video_spots_save_dir, state_path), "rb") as f:
            cell_spot: CellSpot = pickle.load(f)
            cell_spots.append(cell_spot)

    # Load cell tracks
    cell_tracks: list[CellTrack] = []
    video_tracks_save_dir = os.path.join(tracks_dir, video_name)
    for state_path in os.listdir(video_tracks_save_dir):
        with open(os.path.join(video_tracks_save_dir, state_path), "rb") as f:
            cell_track: CellTrack = pickle.load(f)
            cell_tracks.append(cell_track)

    # Detect metaphase spots
    tracks_merging_factory.pre_process_spots(
        cell_tracks,
        cell_spots,
        video,
        metaphase_model_path,
        hmm_metaphase_parameters_file,
        predictions_file,
        video_name,
        only_predictions_update,
    )

    # If the goal is only to update predictions file, stop here
    if only_predictions_update:
        return None

    # Plug tracks occurring at frame>0 to closest metaphase
    mitosis_tracks = tracks_merging_factory.get_tracks_to_merge(cell_tracks)

    print(
        f"\nPredictions performed successfully. {len(mitosis_tracks)} divisions detected."
    )

    # Update useful attributes for each track
    print("Updating division attributes.")
    for i, mitosis_track in enumerate(tqdm(mitosis_tracks)):
        mitosis_track.id = i
        mitosis_track.update_mitosis_start_end(cell_tracks, mitosis_tracks)
        mitosis_track.update_key_events_frame(cell_tracks)
        mitosis_track.update_mitosis_position_contour(cell_tracks)

        # Save mitosis track
        if mitoses_dir is not None and save:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{i}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                mitoses_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)

    # Save updated cell tracks
    if save:
        for cell_track in cell_tracks:
            state_path = f"track_{cell_track.track_id}.bin"
            save_path = os.path.join(
                video_tracks_save_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(cell_track, f)

    return mitosis_tracks, cell_spots, cell_tracks



from __future__ import annotations
from typing import Optional
from tqdm import tqdm
import numpy as np
import pandas as pd

from ..constants.tracking import (
    FRAMES_AROUND_METAPHASE,
    INTERPHASE_INDEX,
    METAPHASE_INDEX,
)
from .track import Track
from .box_dimensions_contour import BoxDimensionsContour
from .box_dimensions import BoxDimensions
from .cell_spot import CellSpot


def get_whole_box_dimensions_advanced(
    tracks: list[CellTrack], frame: int
) -> BoxDimensionsContour:
    """
    Merge different tracks.

    Parameters
    ----------
    predictions : list[int]
        list of predictions for each frame of the track.

    Returns
    -------
    BoxDimensionsContour : Box dimension of merged tracks.

    """
    box_dimensions_contour = BoxDimensionsContour()

    # For all tracks: mother and daughter(s)
    for track in tracks:
        if frame in track.spots:
            current_spot = track.spots[frame]
            box_dimensions_contour.list_points.append(current_spot.spot_points)
            box_dimensions_contour.update(
                current_spot.abs_min_x,
                current_spot.abs_max_x,
                current_spot.abs_min_y,
                current_spot.abs_max_y,
            )
    return box_dimensions_contour


def generate_tracking_movie(
    tracks: list[CellTrack], video: np.ndarray
) -> np.ndarray:
    """
    Generate tracking movie.

    Parameters
    ----------
    tracks : list[CellTrack]
        List of tracks.
    video : np.ndarray
        Video. TYXC.

    Returns
    -------
    np.ndarray
        Tracking movie. TYX.
    """
    nb_frames, height, width = video.shape[:-1]

    mask = np.zeros((nb_frames, width, height)).astype(np.uint8)  # TXY
    for frame in tqdm(range(nb_frames)):
        for track in tracks:
            box_dim_contours = get_whole_box_dimensions_advanced(
                [track], frame
            )  # (x, y)
            mask[frame] = np.maximum(
                mask[frame],
                int(track.track_id + 1)
                * box_dim_contours.get_mask((width, height)),
            )

    # Switch X and Y axes to match video shape
    mask = np.moveaxis(mask, -1, 1)  # TYX
    return mask


class CellTrack(Track[CellSpot]):
    """
    Cell track.
    """

    max_frame_gap = 3

    def __init__(
        self, track_id: int, track_spots_ids: set[int], start: int, stop: int
    ) -> None:
        super().__init__(track_id)
        self.track_spots_ids = track_spots_ids
        self.start = start
        self.stop = stop

        self.metaphase_spots: list[CellSpot] = []

    @classmethod
    def from_spots(cls, track_id: int, spots: list[CellSpot]) -> CellTrack:
        """
        Create a CellTrack from a list of CellSpot.

        Parameters
        ----------
        track_id : int
            Track identifier.
        """
        track_spots_ids = set([spot.id for spot in spots])
        start = min([spot.frame for spot in spots])
        stop = max([spot.frame for spot in spots])
        track = cls(track_id, track_spots_ids, start, stop)
        for spot in spots:
            track.add_spot(spot)
        return track

    def update_metaphase_spots(self, predictions: list[int]) -> None:
        """
        Populate metaphase_spots with candidates.

        Parameters
        ----------
        predictions : list[int]
            list of predictions for each frame of the track.

        Returns
        -------
        None.

        """
        for idx, frame in enumerate(sorted(self.spots.keys())):
            self.spots[frame].predicted_phase = predictions[idx]

        # Store last metaphase spot of each group
        metaphase_finished = False
        for frame in range(self.start, self.stop + 1):
            # Ignore first spots of cell as they are metaphase only if end of previous metaphase
            if predictions[frame - self.start] == INTERPHASE_INDEX:
                metaphase_finished = True
            if not metaphase_finished:
                continue

            # From this point, get metaphase spots
            if (
                frame in self.spots  # current frame contains a spot
                and predictions[frame - self.start]
                == METAPHASE_INDEX  # current spot is in metaphase
                and frame != self.stop  # current spot is not last spot
                and (
                    predictions[frame - self.start + 1] != METAPHASE_INDEX
                    and frame in self.spots
                )  # next frame is not a spot in metaphase
            ):
                self.metaphase_spots.append(self.spots[frame])

    def has_close_metaphase(self, spot: CellSpot, target_frame: int) -> bool:
        """
        Parameters
        ----------
        spot : CellSpot
            Only used to update corresponding metaphase spot.

        Returns
        -------
        bool : True if corresponding track contains one metaphase spot close to target frame.

        """
        # Look for metaphase spot
        for metaphase_spot in self.metaphase_spots:
            if (
                abs(metaphase_spot.frame - target_frame)
                < FRAMES_AROUND_METAPHASE
            ):
                # Mother track found!
                spot.corresponding_metaphase_spot = metaphase_spot
                return True
        return False

    def compute_metaphase_iou(self, daughter_track: CellTrack) -> float:
        """
        Get intersection between daughter area at first frame and self area at previous frame.
        Get self area at previous frame. Returns the quotient.

        Ideally, it should be close to 0.5 as a daughter cell should occupy 50% of the area
        of the mother cell.

        Parameters
        ----------
        daughter_track : CellTrack
            Potential daughter track.

        Returns
        -------
        float
            Intersection Over Union.
        """

        daughter_track_first_frame = min(daughter_track.spots.keys())

        # If current track starts at first frame, ignore as it cannot be a mother track
        if self.start == daughter_track_first_frame:
            return -1

        # Compute two regions
        self_previous_region = self.compute_contour_from_tracks(
            daughter_track_first_frame - 1, relative=False
        )
        daughter_region = daughter_track.compute_contour_from_tracks(
            daughter_track_first_frame, relative=False
        )

        local_shape = (
            max(self_previous_region.max_y, daughter_region.max_y),
            max(self_previous_region.max_x, daughter_region.max_x),
        )

        # Compute masks
        self_previous_region_mask = self_previous_region.get_mask(local_shape)
        daughter_region_mask = daughter_region.get_mask(local_shape)

        # Compute intersection of both regions
        overlap = np.sum(self_previous_region_mask * daughter_region_mask)

        # Compute previous region area
        previous_area = np.sum(self_previous_region_mask)

        return overlap / previous_area

    def compute_contour_from_tracks(
        self,
        frame: int,
        previous_box_dimensions_contour: Optional[BoxDimensionsContour] = None,
        additional_tracks: Optional[list[CellTrack]] = None,
        relative: bool = True,
    ) -> BoxDimensionsContour:
        """
        Compute contours at given frame.

        Parameters
        ----------
        frame : int
            Frame at which contour is computed.
        previous_box_dimensions_contour : Optional[BoxDimensionsContour] = None
            If specified and current track has no spot at frame, no computation is done
            and previous_box_dimensions_contour is returned.
        additional_tracks : Optional[list[CellTrack]] = None
            If specified, perform computation on both current track and these additional tracks.
        relative : bool = True
            If True, indexes are relative to current track position.

        Returns
        -------
        BoxDimensionsContour : Track(s) contour.

        """
        tracks = [self]
        if additional_tracks is not None:
            tracks = tracks + additional_tracks

        box_dimensions_advanced = get_whole_box_dimensions_advanced(
            tracks, frame
        )

        # If missing spot at this frame...
        if box_dimensions_advanced.is_empty():
            # ... use previous frame data if provided
            if previous_box_dimensions_contour:
                return previous_box_dimensions_contour

            # ... or try with previous frame if not
            for _ in range(CellTrack.max_frame_gap):
                frame = frame - 1
                box_dimensions_advanced = get_whole_box_dimensions_advanced(
                    tracks, frame
                )
                if not box_dimensions_advanced.is_empty():
                    break

        # Should not be empty after this loop
        if box_dimensions_advanced.is_empty():
            raise ValueError(
                f"No previous box & Tracks with no spots in {CellTrack.max_frame_gap} frames in a row"
            )

        box_dimensions_advanced.update_list_points(relative)

        return box_dimensions_advanced

    def get_spots_data(
        self, raw_spots: list[CellSpot], raw_video: np.ndarray
    ) -> list[np.array]:
        """
        Generate crops around cells for metaphase CNN inference.

        Parameters
        ----------
        raw_spots : list[CellSpot]
            All video spots.
        raw_video : np.ndarray
            TYXC

        Returns
        -------
        list[np.array]: Cell crops (CYX)

        """

        spot_abs_positions = {}  # {frame: BoxDimensions}
        cell_crops = []  # CYX

        for spot in raw_spots:
            # Ignore spots before or after current track
            if spot.frame < self.start or spot.frame > self.stop:
                continue
            # Ignore spots not in current track
            if spot.id not in self.track_spots_ids:
                continue
            # Store positions
            spot_abs_positions[spot.frame] = BoxDimensions(
                spot.abs_min_x,
                spot.abs_max_x,
                spot.abs_min_y,
                spot.abs_max_y,
            )
            # Store all spots
            spot.track_id = self.track_id  # add track information
            self.spots[spot.frame] = spot

        # If no spot in track for current frame, use previous frame position
        for frame in range(self.start, self.stop + 1):
            if frame in spot_abs_positions:
                min_y, max_y, min_x, max_x = (
                    spot_abs_positions[frame].min_y,
                    spot_abs_positions[frame].max_y,
                    spot_abs_positions[frame].min_x,
                    spot_abs_positions[frame].max_x,
                )
            nucleus = raw_video[frame, min_y:max_y, min_x:max_x, :]  # YXC
            nucleus = np.moveaxis(nucleus, -1, 0)  # CYX
            cell_crops.append(nucleus)

        return cell_crops

    @staticmethod
    def track_df_to_track_list(
        track_df: pd.DataFrame,
        spots: dict[int, list[CellSpot]],
    ) -> list[CellTrack]:

        track_df.reset_index(inplace=True)
        track_df.dropna(inplace=True)
        id_to_track = {}

        for _, row in track_df.iterrows():
            track_id = row["track_id"]
            track: list = id_to_track.get(track_id)
            if track is None:
                id_to_track[track_id] = []
                track = id_to_track[track_id]
            frame = row["frame"]
            idx_in_frame = row["idx_in_frame"]
            track.append(spots[int(frame)][int(idx_in_frame)])
        return [
            CellTrack.from_spots(track_id, spots)
            for track_id, spots in enumerate(id_to_track.values())
        ]



from __future__ import annotations
from enum import IntEnum


class ImpossibleDetection(IntEnum):
    """Enum to represent the different types of impossible detections."""

    NORMAL = 0
    NO_MID_BODY_DETECTED = -1
    MORE_THAN_TWO_DAUGHTER_TRACKS = -2
    NEAR_BORDER = -3
    NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS = -4
    MT_CUT_AT_CYTOKINESIS = -5
    NO_CUT_DETECTED = -6
    LIGHT_SPOT = -7
    TOO_SHORT_CUT = -8
    METAPHASE_AFTER_CYTOKINESIS = -9

    @staticmethod
    def display(detection: ImpossibleDetection) -> bool:
        """Check if the detection status should be displayed.

        Some ImpossibleDetection status are usually linked
        to a false cell division detection, hence are not plotted.

        Parameters
        ----------
        detection : ImpossibleDetection
            Detection status.

        Returns
        -------
        bool
            Whether the mitosis should be displayed.
        """
        return detection in [
            ImpossibleDetection.NORMAL,
            ImpossibleDetection.NEAR_BORDER,
            ImpossibleDetection.NO_CUT_DETECTED,
            ImpossibleDetection.LIGHT_SPOT,
        ]



"""Playground to run mitosis track generation."""

import os
from typing import Optional
from matplotlib import pyplot as plt

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.utils.cell_spot import CellSpot
from cut_detector.utils.cell_track import CellTrack
from cut_detector.utils.mitosis_track import MitosisTrack
from cut_detector.utils.tools import re_organize_channels
from cut_detector.widget_functions.mitosis_track_generation import (
    perform_mitosis_track_generation,
)


def plot_predictions_evolution(
    raw_spots: list[CellSpot],
    raw_tracks: list[CellTrack],
    mitosis_tracks: list[MitosisTrack],
) -> None:
    """
    Plot predictions evolution. Used in playground.

    Parameters
    ----------
    raw_spots : list[CellSpot]
        Raw spots.
    raw_tracks : list[CellTrack]
        Raw tracks.
    mitosis_tracks : list[MitosisTrack]
        Mitosis tracks.
    """
    # Spots detected by Cellpose
    detected_spots = {}
    for raw_spot in raw_spots:
        if raw_spot.frame not in detected_spots:
            detected_spots[raw_spot.frame] = 0
        detected_spots[raw_spot.frame] += 1

    # Mitoses identified by current method, minus ending tracks
    detected_mitoses = {frame: detected_spots[0] for frame in detected_spots}
    max_frame = max(detected_spots.keys())
    for mitosis_track in mitosis_tracks:
        for frame in range(mitosis_track.metaphase_frame, max_frame + 1):
            detected_mitoses[frame] += len(mitosis_track.daughter_track_ids)

    min_frame, max_frame = 0, max(detected_spots.keys())

    detected_tracks = {}
    metaphase_spots = []
    for track in raw_tracks:
        metaphase_spots = metaphase_spots + [
            metaphase_spot.frame for metaphase_spot in track.metaphase_spots
        ]
        for frame in range(track.start, track.stop + 1):
            if frame not in detected_tracks:
                detected_tracks[frame] = 0
            detected_tracks[frame] += 1

    data_spots = [detected_spots[0]] + [
        detected_spots[i] if i in detected_spots else 0
        for i in range(min_frame, max_frame)
    ]
    data_tracks = [detected_tracks[0]] + [
        detected_tracks[i] if i in detected_tracks else 0
        for i in range(min_frame, max_frame)
    ]
    data_mitoses = [detected_mitoses[0]] + [
        detected_mitoses[i] if i in detected_mitoses else 0
        for i in range(min_frame, max_frame)
    ]

    _, ax = plt.subplots(1, 1, figsize=(15, 5))
    ax.step(
        list(range(min_frame, max_frame + 1)), data_spots, "r", linewidth=8.0
    )
    ax.step(
        list(range(min_frame, max_frame + 1)), data_tracks, "g", linewidth=8.0
    )
    ax.step(
        list(range(min_frame, max_frame + 1)), data_mitoses, "b", linewidth=8.0
    )

    ending_tracks = [track.stop + 1 for track in raw_tracks]

    # Plot first for legend
    ax.axvline(-10, color="y", linewidth=2.0)
    ax.axvline(-10, color="c", linewidth=2.0)
    ax.axvline(-10, color="k", linewidth=2.0, linestyle="--")

    # Potential mitoses
    metaphase_spots = list(set(metaphase_spots))
    for metaphase_spot_frame in metaphase_spots:
        ax.axvline(metaphase_spot_frame, color="y", linewidth=2.0)

    # Actual mitoses
    actual_mitoses = [mitosis.metaphase_frame for mitosis in mitosis_tracks]
    for actual_mitosis in actual_mitoses:
        ax.axvline(actual_mitosis, color="c", linewidth=2.0)

    # Ending tracks
    for ending_track_frame in ending_tracks:
        ax.axvline(
            ending_track_frame, color="k", linewidth=2.0, linestyle="--"
        )

    ax.set_xlim(min_frame, max_frame)

    ax.legend(
        [
            "Spots",
            "Tracks",
            "Mitoses evolution",
            "Potential mitoses",
            "Actual mitoses",
            "Ending tracks",
        ],
        loc="best",
    )
    plt.show()


def main(
    image_path: Optional[str] = get_data_path("videos"),
    video_name: Optional[str] = "example_video",
    spots_dir: Optional[str] = get_data_path("spots"),
    tracks_dir: Optional[str] = get_data_path("tracks"),
):
    """
    Parameters
    ----------
    image_path : str
        Path to the image to process.
    video_name : str
        Video name.
    spots_dir : str
        Path to the spots directory.
    tracks_dir : str
        Path to the tracks directory.
    """
    # If image_path or model_path are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])

    # Read image and preprocess if needed
    image = TiffReader(image_path).image  # TCZYX
    image = re_organize_channels(image.squeeze())  # TYXC

    mitosis_tracks, cell_spots, cell_tracks = perform_mitosis_track_generation(
        image, video_name, spots_dir, tracks_dir
    )

    plot_predictions_evolution(cell_spots, cell_tracks, mitosis_tracks)


if __name__ == "__main__":
    main()



import os
from skimage import io

from cut_detector._widget import segmentation_tracking
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.segmentation_tracking import (
    perform_tracking,
)


def test_open_track_generation_widget():
    """Test opening the track generation widget."""
    segmentation_tracking()


def test_track_generation():
    """Test track generation."""
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")
    )  # TYXC

    cell_spots, cell_tracks, _ = perform_tracking(video, save=False)

    assert 230 <= len(cell_spots) <= 240  # should be 236
    assert len(cell_tracks) == 5



import os
from skimage import io

from cut_detector._widget import results_saving
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.save_results import perform_results_saving


def test_open_results_saving_widget():
    """Test opening the results saving widget."""
    results_saving()


def test_results_saving():
    """Test results saving."""
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")
    )  # TYXC

    perform_results_saving(
        get_data_path("mitoses"),
        show=False,
        save_dir=get_data_path("results"),
        verbose=True,
        video=video,
    )



from cut_detector.utils.tools import upload_annotations_folder

if __name__ == "__main__":

    ANNOTATIONS_FOLDER = ""
    VIDEO_FOLDER = ""
    MITOSES_FOLDER = ""

    upload_annotations_folder(
        ANNOTATIONS_FOLDER, VIDEO_FOLDER, MITOSES_FOLDER, save=True
    )



import torch
import torch.nn as nn
from torch import optim

from cnn_framework.utils.data_loader_generators.classifier_data_loader_generator import (
    ClassifierDataLoaderGenerator,
)
from cnn_framework.utils.model_managers.cnn_model_manager import CnnModelManager
from cnn_framework.utils.metrics.classification_accuracy import ClassificationAccuracy
from cnn_framework.utils.parsers.cnn_parser import CnnParser
from cnn_framework.utils.models.resnet_classifier import ResnetClassifier

from binary_bridges_cnn.data_set import BinaryBridgesCnnDataSet
from binary_bridges_cnn.model_params import BinaryBridgesModelParams


def main(params):
    loader_generator = ClassifierDataLoaderGenerator(params, BinaryBridgesCnnDataSet)
    train_dl, val_dl, test_dl = loader_generator.generate_data_loader()

    # Load pretrained model
    model = ResnetClassifier(
        nb_classes=params.nb_classes,  # output classes
        nb_input_channels=len(params.c_indexes) * len(params.z_indexes),
        encoder_name=params.encoder_name,
    )

    manager = CnnModelManager(model, params, ClassificationAccuracy)

    optimizer = optim.Adam(
        model.parameters(),
        lr=float(params.learning_rate),
        betas=(params.beta1, params.beta2),
    )  # define the optimization
    loss_function = nn.CrossEntropyLoss()  # define the loss function
    manager.fit(train_dl, val_dl, optimizer, loss_function)

    scores = []
    for model_path, name in zip(
        [manager.model_save_path_early_stopping, manager.model_save_path],
        ["early stopping", "final"],
    ):
        print(f"\nPredicting with {name} model.")
        # Update model with saved one
        manager.model.load_state_dict(torch.load(model_path))
        manager.predict(test_dl)
        score = manager.training_information.score
        scores.append(score)

    manager.training_information.score = max(scores)
    manager.write_useful_information()


if __name__ == "__main__":
    parser = CnnParser()
    args = parser.arguments_parser.parse_args()

    parameters = BinaryBridgesModelParams()
    parameters.update(args)

    main(parameters)



from cut_detector.utils.mid_body_spot import MidBodySpot
from cut_detector.utils.mid_body_track import MidBodyTrack


def test_mid_body_track_fill_gaps_standard():
    """Test the fill_gaps method of the MidBodyTrack class."""
    track = MidBodyTrack(0)
    track.add_spot(MidBodySpot(0, 0, 0, 0.0, 0.0, 0.0, 0.0))
    track.add_spot(MidBodySpot(2, 2, 2, 2.0, 2.0, 2.0, 2.0))
    track.add_spot(MidBodySpot(3, 3, 3, 3.0, 3.0, 3.0, 3.0))
    track.fill_gaps()
    assert track.spots[1].x == 1


def test_mid_body_track_fill_gaps_none():
    """Same with missing attributes."""
    track = MidBodyTrack(0)
    track.add_spot(MidBodySpot(0, 0, 0, 0.0))
    track.add_spot(MidBodySpot(2, 2, 2, 2.0))
    track.fill_gaps()
    assert track.spots[1].x == 1
    assert track.spots[1].area is None


def test_mid_body_track_fill_gaps_bigger():
    """Same with bigger gap."""
    track = MidBodyTrack(0)
    track.add_spot(MidBodySpot(2, 2, 2, 2.0))
    track.add_spot(MidBodySpot(5, 5, 5, 5.0))
    track.fill_gaps()
    assert track.spots[3].x == 3
    assert track.spots[4].x == 4


def test_mid_body_track_fill_gaps_all():
    """Same with all previous conditions."""
    track = MidBodyTrack(0)
    track.add_spot(MidBodySpot(0, 0, 0, 0.0))
    track.add_spot(MidBodySpot(2, 2, 2, 2.0))
    track.add_spot(MidBodySpot(5, 5, 5, 5.0))
    track.fill_gaps()
    assert track.spots[1].x == 1
    assert track.spots[3].x == 3
    assert track.spots[4].x == 4



import os
from typing import Optional
import pickle
from skimage import io
import napari

from cut_detector.data.tools import get_data_path
from cut_detector.factories.results_saving_factory import ResultsSavingFactory
from cut_detector.utils.cell_track import CellTrack


def main(
    image_path: Optional[str] = os.path.join(
        get_data_path("videos"), "example_video.tif"
    ),
    mitoses_path: Optional[str] = get_data_path("mitoses"),
    segmentation_results_path: Optional[str] = os.path.join(
        get_data_path("segmentation_results"), "example_video.bin"
    ),
    tracks_dir: Optional[str] = get_data_path("tracks"),
    video_name="example_video",
) -> None:
    """
    Parameters
    ----------
    image_path : str
        Path to the image to display, by default example_video.tif.
    mitoses_path : str
        Path to the mitoses, by default mitoses.
    segmentation_results_path : str
        Path to the segmentation results, by default example_video.bin.
    """

    # Create a Napari viewer
    viewer = napari.Viewer()

    # Load Cellpose results
    with open(segmentation_results_path, "rb") as f:
        cellpose_results = pickle.load(f)  # TYX

    # Load cell tracks
    cell_tracks: list[CellTrack] = []
    video_tracks_save_dir = os.path.join(tracks_dir, video_name)
    for state_path in os.listdir(video_tracks_save_dir):
        with open(os.path.join(video_tracks_save_dir, state_path), "rb") as f:
            cell_track: CellTrack = pickle.load(f)
            cell_tracks.append(cell_track)

    # Add video
    video = io.imread(image_path)  # TYXC

    # Match Napari video display
    viewer.add_image(video, name="example_video", rgb=True)

    # Load mitosis tracks
    mitosis_tracks = []
    for state_path in os.listdir(mitoses_path):
        with open(os.path.join(mitoses_path, state_path), "rb") as f:
            mitosis_track = pickle.load(f)
        mitosis_tracks.append(mitosis_track)

    ResultsSavingFactory().generate_napari_tracking_mask(
        mitosis_tracks,
        video,
        viewer,
        segmentation_results=cellpose_results,
        cell_tracks=cell_tracks,
    )

    napari.run()


if __name__ == "__main__":
    main()



""" A modified version of LapTrack.
- Maximum distance is only applied to the euclidian distance
"""

import numpy as np
from functools import partial
from typing import Callable, List, cast, Union
from pydantic import Field
from scipy.spatial.distance import cdist
from laptrack import LapTrack, ParallelBackend
import networkx as nx

from laptrack._typing_utils import EdgeType
from laptrack._coo_matrix_builder import coo_matrix_builder
from laptrack._optimization import lap_optimization
from laptrack._cost_matrix import build_frame_cost_matrix


class SpatialLapTrack(LapTrack):
    """
    SpatialLapTrack is a modification of LapTrack where
    track_cost_cutoff and gap_closing_cost_cutoff are compared
    against euclidian spatial distance only.

    (splitting and merging cost cutoff works as usual because
    they are not used in the project for now)

    This is only for filtering out values:
    Optimisation is still run against the chosen metric.
    """

    spatial_coord_slice: slice = Field(
        ..., description="A slice that is used to subset coords"
    )

    spatial_metric: Union[str, Callable] = Field(
        "euclidean",
        description="The metric to use to compute spatial distances",
    )

    class Config:
        arbitrary_types_allowed = True

    def _predict_links(
        self, coords, segment_connected_edges, split_merge_edges
    ) -> nx.Graph:
        """
        Link particles between frames according to the cost function.

        Parameters
        ----------
            coords : List[np.ndarray]
                the input coordinates
            segment_connected_edges : EdgesType
                the connected edges list that will be connected in this step
            split_merge_edges : EdgesType
                the connected edges list that will be connected in split and merge step

        Returns
        -------
            nx.Graph: The resulted tree.

        """
        # initialize tree
        track_tree = nx.Graph()
        for frame, coord in enumerate(coords):
            for j in range(coord.shape[0]):
                track_tree.add_node((frame, j))

        # linking between frames
        edges_list = list(segment_connected_edges) + list(split_merge_edges)

        def _predict_link_single_frame(
            frame: int,
            coord1: np.ndarray,
            coord2: np.ndarray,
        ) -> List[EdgeType]:
            force_end_indices = [
                e[0][1] for e in edges_list if e[0][0] == frame
            ]
            force_start_indices = [
                e[1][1] for e in edges_list if e[1][0] == frame + 1
            ]

            dist_matrix = cdist(coord1, coord2, metric=self.track_dist_metric)
            dist_matrix[force_end_indices, :] = np.inf
            dist_matrix[:, force_start_indices] = np.inf

            # While this has not been checked, it is highly likely
            # that coord are ordered the way specified in predict_dataframe
            spatial_coord1 = coord1[:, self.spatial_coord_slice]
            spatial_coord2 = coord2[:, self.spatial_coord_slice]

            spatial_dist_matrix = cdist(
                spatial_coord1, spatial_coord2, metric=self.spatial_metric
            )
            spatial_dist_matrix[force_end_indices, :] = np.inf
            spatial_dist_matrix[:, force_start_indices] = np.inf

            # ind = np.where(dist_matrix < self.track_cost_cutoff)
            ind = np.where(spatial_dist_matrix < self.track_cost_cutoff)
            dist_matrix = coo_matrix_builder(  # keeping the true score here
                dist_matrix.shape,
                row=ind[0],
                col=ind[1],
                data=dist_matrix[(*ind,)],
                dtype=dist_matrix.dtype,
            )

            cost_matrix = build_frame_cost_matrix(
                dist_matrix,
                track_start_cost=self.track_start_cost,
                track_end_cost=self.track_end_cost,
            )
            xs, _ = lap_optimization(cost_matrix)

            count1 = dist_matrix.shape[0]
            count2 = dist_matrix.shape[1]
            connections = [(i, xs[i]) for i in range(count1) if xs[i] < count2]
            edges: List[EdgeType] = [
                ((frame, connection[0]), (frame + 1, connection[1]))
                for connection in connections
            ]
            return edges

        if self.parallel_backend == ParallelBackend.serial:
            all_edges = []
            for frame, (coord1, coord2) in enumerate(
                zip(coords[:-1], coords[1:])
            ):
                edges = _predict_link_single_frame(frame, coord1, coord2)
                all_edges.extend(edges)
        elif self.parallel_backend == ParallelBackend.ray:
            try:
                import ray  # type: ignore (removes the import warning)
            except ImportError:
                raise ImportError(
                    "Please install `ray` to use `ParallelBackend.ray`."
                )
            remote_func = ray.remote(_predict_link_single_frame)
            res = [
                remote_func.remote(frame, coord1, coord2)
                for frame, (coord1, coord2) in enumerate(
                    zip(coords[:-1], coords[1:])
                )
            ]
            all_edges = sum(ray.get(res), [])
        else:
            raise ValueError(
                f"Unknown parallel backend {self.parallel_backend}. "
                + f"Must be one of {', '.join([ps.name for ps in ParallelBackend])}."
            )

        track_tree.add_edges_from(all_edges)
        track_tree.add_edges_from(segment_connected_edges)
        return track_tree

    def _get_gap_closing_matrix(
        self, segments_df, *, force_end_nodes=[], force_start_nodes=[]
    ):
        """
        Generate the cost matrix for connecting segment ends.

        Parameters
        ----------
        segments_df : pd.DataFrame
            must have the columns "first_frame", "first_index", "first_crame_coords", "last_frame", "last_index", "last_frame_coords"
        force_end_nodes : list of int
            the indices of the segments_df that is forced to be end for future connection
        force_start_nodes : list of int
            the indices of the segments_df that is forced to be start for future connection

        Returns
        -------
        segments_df: pd.DataFrame
            the segments dataframe with additional column "gap_closing_candidates"
            (index of the candidate row of segments_df, the associated costs)
         gap_closing_dist_matrix: coo_matrix_builder
            the cost matrix for gap closing candidates

        """
        if self.gap_closing_cost_cutoff:

            def to_gap_closing_candidates(row, segments_df):
                # if the index is in force_end_indices, do not add to gap closing candidates
                if (row["last_frame"], row["last_index"]) in force_end_nodes:
                    return [], []

                target_coord = row["last_frame_coords"]
                target_spatial_coord = target_coord[self.spatial_coord_slice]
                frame_diff = segments_df["first_frame"] - row["last_frame"]

                # only take the elements that are within the frame difference range.
                # segments in df is later than the candidate segment (row)
                indices = (1 <= frame_diff) & (
                    frame_diff <= self.gap_closing_max_frame_count
                )
                df = segments_df[indices]
                force_start = df.apply(
                    lambda row: (row["first_frame"], row["first_index"])
                    in force_start_nodes,
                    axis=1,
                )
                df = df[~force_start]
                # do not connect to the segments that is forced to be start
                # note: can use KDTree if metric is distance,
                # but might not be appropriate for general metrics
                # https://stackoverflow.com/questions/35459306/find-points-within-cutoff-distance-of-other-points-with-scipy # noqa
                # TrackMate also uses this (trivial) implementation.
                if len(df) > 0:
                    target_dist_matrix = cdist(
                        [target_coord],
                        np.stack(df["first_frame_coords"].values),
                        metric=self.gap_closing_dist_metric,
                    )
                    assert target_dist_matrix.shape[0] == 1

                    # spatial distance
                    spatial_target_dist_matrix = cdist(
                        [target_spatial_coord],
                        # np.stack([df["first_frame_coords"].values[0][self.spatial_coord_slice]]),
                        np.stack(df["first_frame_coords"].values)[
                            :, self.spatial_coord_slice
                        ],
                        metric=self.spatial_metric,
                    )
                    assert spatial_target_dist_matrix.shape[0] == 1

                    indices2 = np.where(
                        # target_dist_matrix[0] < self.gap_closing_cost_cutoff
                        spatial_target_dist_matrix[0]
                        < self.gap_closing_cost_cutoff
                    )[0]
                    return (
                        df.index[indices2].values,
                        target_dist_matrix[0][indices2],
                    )
                else:
                    return [], []

            if self.parallel_backend == ParallelBackend.serial:
                segments_df["gap_closing_candidates"] = segments_df.apply(
                    partial(
                        to_gap_closing_candidates, segments_df=segments_df
                    ),
                    axis=1,
                )
            elif self.parallel_backend == ParallelBackend.ray:
                try:
                    import ray  # type: ignore
                except ImportError:
                    raise ImportError(
                        "Please install `ray` to use `ParallelBackend.ray`."
                    )
                remote_func = ray.remote(to_gap_closing_candidates)
                segments_df_id = ray.put(segments_df)
                res = [
                    remote_func.remote(row, segments_df_id)
                    for _, row in segments_df.iterrows()
                ]
                segments_df["gap_closing_candidates"] = ray.get(res)
            else:
                raise ValueError(
                    f"Unknown parallel_backend {self.parallel_backend}. "
                )
        else:
            segments_df["gap_closing_candidates"] = [([], [])] * len(
                segments_df
            )

        N_segments = len(segments_df)
        gap_closing_dist_matrix = coo_matrix_builder(
            (N_segments, N_segments), dtype=np.float32
        )
        for ind, row in segments_df.iterrows():
            candidate_inds = row["gap_closing_candidates"][0]
            candidate_costs = row["gap_closing_candidates"][1]
            # row ... track end, col ... track start
            gap_closing_dist_matrix[(int(cast(int, ind)), candidate_inds)] = (
                candidate_costs
            )

        return segments_df, gap_closing_dist_matrix



import os
import pickle
from typing import Optional
import numpy as np
from ..utils.cell_spot import CellSpot
from ..utils.cell_track import CellTrack
from ..factories.segmentation_tracking_factory import (
    SegmentationTrackingFactory,
)
from ..models.tools import get_model_path


def perform_tracking(
    video: np.ndarray,
    model_path: Optional[str] = None,
    video_name: Optional[str] = None,
    spots_save_dir: Optional[str] = None,
    tracks_save_dir: Optional[str] = None,
    save: bool = True,
) -> tuple[list[CellSpot], list[CellTrack], np.ndarray]:
    """
    Run cell segmentation and tracking on the given video.

    Parameters
    ----------
    video : np.ndarray, TYXC
        The video to run the segmentation and tracking on.
    model_path : str, optional
        Path to the cellpose model, by default None.
    video_name : str, optional
        Name of the video, by default None.
    spots_save_dir : str, optional
        Directory to save cell spots, by default None.
    tracks_save_dir : str, optional
        Directory to save cell tracks, by default None.
    save : bool, optional
        Save the results, by default True.

    Returns
    -------
    tuple[list[CellSpot], list[CellTrack], np.ndarray]
        List of cell spots, list of cell tracks, segmentation results.
    """

    print("### SEGMENTATION & TRACKING ###")

    if model_path is None:
        model_path = os.path.join(
            get_model_path("segmentation"), "segmentation_model"
        )

    segmentation_tracking_factory = SegmentationTrackingFactory(model_path)
    # Swap dimensions from TYXC to TCYX
    video = np.moveaxis(video, 3, 1)
    cell_spots, cell_tracks, segmentation_results = (
        segmentation_tracking_factory.perform_segmentation_tracking(video)
    )

    if not save:
        return cell_spots, cell_tracks, segmentation_results

    assert spots_save_dir is not None
    assert tracks_save_dir is not None

    # Create saving directories if they do not exist
    video_spots_save_dir = os.path.join(spots_save_dir, video_name)
    if not os.path.exists(video_spots_save_dir):
        os.makedirs(video_spots_save_dir)
    video_tracks_save_dir = os.path.join(tracks_save_dir, video_name)
    if not os.path.exists(video_tracks_save_dir):
        os.makedirs(video_tracks_save_dir)

    # Save cell spots
    for cell_spot in cell_spots:
        state_path = f"spot_{cell_spot.id}.bin"
        save_path = os.path.join(
            video_spots_save_dir,
            state_path,
        )
        with open(save_path, "wb") as f:
            pickle.dump(cell_spot, f)

    # Save cell tracks
    for cell_track in cell_tracks:
        state_path = f"track_{cell_track.track_id}.bin"
        save_path = os.path.join(
            video_tracks_save_dir,
            state_path,
        )
        with open(save_path, "wb") as f:
            pickle.dump(cell_track, f)

    return cell_spots, cell_tracks, segmentation_results



from cnn_framework.utils.model_params.base_model_params import BaseModelParams
from cnn_framework.utils.dimensions import Dimensions


class BinaryBridgesModelParams(BaseModelParams):
    """
    Bridges CNN model params.
    """

    def __init__(self, inference_mode=False):
        super().__init__("binary_bridges_cnn")

        self.input_dimensions = Dimensions(height=100, width=100)

        self.num_epochs = 100
        self.learning_rate = 1e-4
        self.c_indexes = [0]
        self.mode = "microtubules"
        self.encoder_name = "resnet18"
        self.num_epochs = 100
        self.learning_rate = 1e-4

        if inference_mode:
            self.class_names = ["2 MT", "1 MT", "0 MT"]
        else:
            self.class_names = ["No micro-tubules", "Micro-tubules"]

        self.nb_classes = len(self.class_names)
        self.hmm_bridges_parameters_file = ""

    def update(self, args=None):
        super().update(args)

        if args.hmm_bridges_parameters_file:
            self.hmm_bridges_parameters_file = args.hmm_bridges_parameters_file
